text \<open>Trivialities\<close>
theory Basic_Utils
  imports QTRS.Trs
begin

lemma finite_check:
  fixes M
  assumes "S \<subset> {x :: nat. x \<le> M}"
  shows "finite S"
  using assms Collect_cong atMost_def finite_atMost finite_subset less_le by fastforce

lemma finite_Max:
  fixes S :: "nat set"
  assumes "finite S" and "Max S > 0" and "S \<noteq> {}"
  shows "Max S \<in> S"
proof (rule ccontr)
  assume A: "(Max S) \<notin> S"
  have C: "\<forall>x \<in> S. x \<le> Max S - 1" using assms A by force
  then have "Max S - 1 < Max S" using assms(2) by linarith
  then have "\<exists>x \<in> S. x > Max S - 1" using Max_gr_iff assms(1) assms(3) by auto
  then show  False using C by auto
qed

lemma finite_nat_subset:
  assumes "S \<subseteq>  {x :: nat. x \<le> M}"
  shows "finite S"
  using assms
  by (simp add:  finite_Collect_le_nat finite_subset)

lemma max_attain:
fixes x s assumes "finite S" "x\<in> S" and "x \<ge> Max S" shows "x = Max S"
proof (rule ccontr)
  assume A:"x \<noteq> Max S"
  have loc: "x > Max S" using assms(3) A by auto
  moreover have "x \<le> Max S" using Max_ge assms(1) assms(2) by auto
  then show False using loc by simp
qed

lemma subs:
  fixes f M
  assumes "\<forall>m n :: nat. n < m \<longrightarrow> f n \<le> f m"
    and  "\<forall>n. f n \<le> (M :: nat)"
  shows "\<exists>N. \<forall>n. n \<ge> N \<longrightarrow> f n = Max (range f)"
proof -
  have "\<forall>x \<in> range f. x \<le> M"
    using assms(2) by auto
  then have "range f \<subseteq> {x :: nat. x \<le> M}"
    by auto
  then have fin:"finite (range f)" using finite_nat_subset by auto
  show ?thesis
  proof (cases "Max (range f)")
    case  0
    show ?thesis using "0"
      by (metis Max.boundedE UNIV_not_empty fin image_is_empty le_0_eq rangeI)
  next
    case (Suc m)
    have "Max (range f) \<in> range f"  using fin  finite_Max by auto
    then obtain N where loc1: "f N = Max (range f)" by auto
    then have  loc2: "\<forall>n. n > N \<longrightarrow> f n \<ge> Max (range f)" using assms(1) by fastforce
    moreover have "\<forall>n. n = N \<longrightarrow> f n \<ge> Max (range f)" using loc1 by auto
    then have "\<forall>n. n \<ge> N \<longrightarrow> f n \<ge> Max (range f)" using loc2
      by (simp add: dual_order.order_iff_strict)
    then show ?thesis  using max_attain fin rangeI by blast
  qed
qed

lemma inc_set_fixpoint:
  fixes f S
  assumes "\<forall>m n :: nat. n < m \<longrightarrow> f n \<subseteq> f m"
  and " \<forall>n. f n \<subseteq> S \<and> finite S"
shows "\<exists>n. \<forall>m. m > n \<longrightarrow> f m = f n"
  using assms 
proof -
  let ?M = "card S"
  have ub:"\<forall>n. (card \<circ> f) n \<le> ?M"
    using assms(2) Finite_Set.card_mono by auto
  have "\<forall>m n. n < m \<longrightarrow> (card \<circ> f) n \<le> (card \<circ> f) m"
    using assms(1) by (metis assms(2) card_mono comp_apply finite_subset)
  then have "\<exists>N. \<forall>n. n \<ge> N \<longrightarrow> (card \<circ> f) n = Max (range (card \<circ> f))"
    using ub subs [of "card \<circ> f"] by auto
  then obtain N where obt_loc:" \<forall>n. n \<ge> N \<longrightarrow> ((card\<circ>f) n) = Max (range (card \<circ> f))"
    by auto
  then have "\<forall>n. n \<ge> N \<longrightarrow> card (f n) = Max (range (card \<circ> f))"
    using comp_def by auto
  then have loc: "\<forall>n. n > N \<longrightarrow> f N \<subseteq> f n" using assms(1) by auto
  also have "\<forall>n. n = N \<longrightarrow> f N \<subseteq> f n" by auto
  then have "\<forall>n. n \<ge> N \<longrightarrow> f N \<subseteq> f n"
    using loc by (simp add: dual_order.order_iff_strict)
  then have "\<forall>n. n \<ge> N \<longrightarrow> card (f N) = card (f n)"
    using obt_loc by auto
  then have "\<forall>n. n \<ge> N \<longrightarrow> f N = f n" using ub Finite_Set.card_subset_eq
    by (metis \<open>\<forall>n \<ge> N. f N \<subseteq> f n\<close> assms(2) rev_finite_subset)
   then show ?thesis using less_imp_le_nat by blast
qed

lemma inc_set_fixpoint':
  fixes f S
  assumes "\<forall>m n :: nat. n \<le> m \<longrightarrow> f n \<subseteq> f m"
  and " \<forall>n. f n \<subseteq> S \<and> finite S"
shows "\<exists>n. \<forall>m. m \<ge> n \<longrightarrow> f m = f n"
  using assms inc_set_fixpoint[of f S] le_neq_implies_less
  by (metis less_or_eq_imp_le)

lemma inc_set_fixpoint2:
  fixes f g S
  assumes "\<forall>m n :: nat. n \<le> m \<longrightarrow> f n \<subseteq> f m \<and> g n \<subseteq> g m"
      and "\<forall>n. f n \<subseteq> S \<and> g n \<subseteq> S"
      and "finite S"
    shows  "\<exists>n. \<forall>m \<ge> n. f m = f n \<and> g m = g n"
proof -
  obtain n\<^sub>1 n\<^sub>2 where fixpoints:
    "\<forall>m. m \<ge> n\<^sub>1 \<longrightarrow> f m = f n\<^sub>1" "\<forall>m. m \<ge> n\<^sub>2 \<longrightarrow> g m = g n\<^sub>2"
    using inc_set_fixpoint'[of f S] inc_set_fixpoint'[of g S] assms by blast+
  define n where "n = max n\<^sub>1 n\<^sub>2"
  have "\<forall>m \<ge> n. f m = f n \<and> g m = g n" using fixpoints n_def
    by (metis max.boundedE max.cobounded1 max.cobounded2)
  thus ?thesis by blast
qed

lemma construct_exI':
  assumes "\<forall>i < n. \<exists>x. P x i"
  shows "\<exists>xs. length xs = n \<and> (\<forall>i < n. P (xs ! i) i)"
  using assms
proof (induction n)
  case (Suc n)
  then obtain x where x: "P x n" by blast
  obtain xs where "length xs = n \<and> (\<forall>i < n. P (xs ! i) i)" using Suc by auto
  hence "length (xs @ [x]) = Suc n \<and> (\<forall>i < Suc n. P ((xs @ [x]) ! i) i)"
    using x by (simp add: less_Suc_eq nth_append)
  thus ?case by blast
qed simp

lemma construct_exI'_double:
  assumes "\<forall>i < n. \<exists>x y. P x y i"
  shows "\<exists>xs ys. length xs = n \<and> length ys = n \<and> (\<forall>i < n. P (xs ! i) (ys!i) i)"
  using assms
proof (induction n)
  case (Suc n)
  then obtain x y where x_y: "P x y n" by blast
  obtain xs ys where "length xs = n \<and> length ys = n \<and> (\<forall>i<n. P (xs ! i) (ys ! i) i)" using Suc by auto
  hence "length (xs @ [x]) = Suc n \<and> length (ys @ [y]) = Suc n \<and>
    (\<forall>i<Suc n. P ((xs @ [x]) ! i) ((ys @ [y]) ! i) i)"
    using x_y by (simp add: less_Suc_eq nth_append)
  thus ?case by blast
qed simp

(* TODO: check whether this is still needed *)
lemma construct_exI:
  assumes "\<forall>i < n. \<exists>xs. P (xs ! i) i"
  shows "\<exists>xs. length xs = n \<and> (\<forall>i < n. P (xs ! i) i)"
  using assms
proof (induction n)
  case (Suc n)
  then obtain ys where ys_n: "P (ys ! n) n" by blast
  obtain xs where "length xs = n \<and> (\<forall>i < n. P (xs ! i) i)" using Suc by auto
  hence "length (xs @ [ys ! n]) = Suc n \<and> (\<forall>i < Suc n. P ((xs @ [ys ! n]) ! i) i)"
    using ys_n by (simp add: less_Suc_eq nth_append)
  thus ?case by blast
qed simp

definition right_ground_trs :: "('f, 'v) trs \<Rightarrow> bool" where
  "right_ground_trs R \<equiv> \<forall>(l, r) \<in> R. ground r"

definition ground_trs :: "('f, 'v) trs \<Rightarrow> bool" where
  "ground_trs R \<equiv> \<forall>(l, r) \<in> R. ground l \<and> ground r"

lemma ground_imp_right_ground_trs [simp]:
  "ground_trs R \<Longrightarrow> right_ground_trs R"
  unfolding ground_trs_def right_ground_trs_def by fast

definition set_to_list :: "'a set \<Rightarrow> 'a list"
  where "set_to_list s = (SOME l. set l = s)"

lemma set_set_to_list:
  "finite s \<Longrightarrow> set (set_to_list s) = s"
  unfolding set_to_list_def by (metis (mono_tags) finite_list some_eq_ex)

end
