theory GTT_Compose_Impl
  imports GTT_Compose GTT_Impl Horn_List
begin

inductive_set \<Delta>\<^sub>\<epsilon>' :: "('p, 'f) ta \<Rightarrow> ('q, 'f) ta \<Rightarrow> ('p \<times> 'q) set" for A B where
  \<Delta>\<^sub>\<epsilon>'_cong: "f ps \<rightarrow> p \<in> ta_rules A \<Longrightarrow> f qs \<rightarrow> q \<in> ta_rules B \<Longrightarrow> length ps = length qs \<Longrightarrow>
   (\<And>i. i < length qs \<Longrightarrow> (ps ! i, qs ! i) \<in> \<Delta>\<^sub>\<epsilon>' A B) \<Longrightarrow> (p, q) \<in> \<Delta>\<^sub>\<epsilon>' A B"
| \<Delta>\<^sub>\<epsilon>'_eps1: "(p, p') \<in> ta_eps A \<Longrightarrow> (p, q) \<in> \<Delta>\<^sub>\<epsilon>' A B \<Longrightarrow> (p', q) \<in> \<Delta>\<^sub>\<epsilon>' A B"
| \<Delta>\<^sub>\<epsilon>'_eps2: "(q, q') \<in> ta_eps B \<Longrightarrow> (p, q) \<in> \<Delta>\<^sub>\<epsilon>' A B \<Longrightarrow> (p, q') \<in> \<Delta>\<^sub>\<epsilon>' A B"

lemma \<Delta>\<^sub>\<epsilon>_eq_\<Delta>\<^sub>\<epsilon>' [code_unfold]:
  "\<Delta>\<^sub>\<epsilon> A B = \<Delta>\<^sub>\<epsilon>' A B"
proof (intro set_eqI iffI, goal_cases lr rl)
  case (lr x) obtain p q where x [simp]: "x = (p, q)" by (cases x)
  obtain t where "ground t" "p \<in> ta_res A t" "q \<in> ta_res B t" using lr by (auto simp: \<Delta>\<^sub>\<epsilon>_def)
  then show ?case unfolding x
  proof (induct t arbitrary: p q)
    case (Fun f ts)
    obtain p' ps where p': "f ps \<rightarrow> p' \<in> ta_rules A" "(p', p) \<in> (ta_eps A)\<^sup>*" "length ps = length ts"
      "\<And>i. i < length ts \<Longrightarrow> ps ! i \<in> ta_res A (ts ! i)" using Fun(3) by auto
    obtain q' qs where q': "f qs \<rightarrow> q' \<in> ta_rules B" "(q', q) \<in> (ta_eps B)\<^sup>*" "length qs = length ts"
      "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res B (ts ! i)" using Fun(4) by auto
    have "(p', q') \<in> \<Delta>\<^sub>\<epsilon>' A B"
      using Fun(1)[OF nth_mem _ p'(4) q'(4)] Fun(2) \<Delta>\<^sub>\<epsilon>'_cong[OF p'(1) q'(1)] p'(3) q'(3) by auto
    with p'(2) have "(p, q') \<in> \<Delta>\<^sub>\<epsilon>' A B" by (induct) (auto intro: \<Delta>\<^sub>\<epsilon>'_eps1)
    with q'(2) show "(p, q)  \<in> \<Delta>\<^sub>\<epsilon>' A B" by (induct) (auto intro: \<Delta>\<^sub>\<epsilon>'_eps2)
  qed auto
next
  case (rl x) obtain p q where x [simp]: "x = (p, q)" by (cases x)
  have "\<exists>t. ground t \<and> p \<in> ta_res A t \<and> q \<in> ta_res B t" using rl unfolding x
  proof (induct)
    case (\<Delta>\<^sub>\<epsilon>'_cong f ps p qs q)
    obtain ts where ts: "ground (ts i) \<and> ps ! i \<in> ta_res A (ts i) \<and> qs ! i \<in> ta_res B (ts i)"
      if "i < length qs" for i using \<Delta>\<^sub>\<epsilon>'_cong(5) by metis
    then show ?case using \<Delta>\<^sub>\<epsilon>'_cong(1-3)
      by (auto intro!: exI[of _ "Fun f (map ts [0..<length qs])"]) blast+
  qed (auto intro: ta_res_eps)
  then show ?case by (auto simp: \<Delta>\<^sub>\<epsilon>_def)
qed

definition \<Delta>\<^sub>\<epsilon>'_rules :: "('p, 'f) ta \<Rightarrow> ('q, 'f) ta \<Rightarrow> ('p \<times> 'q) horn set" where
  "\<Delta>\<^sub>\<epsilon>'_rules A B =
    {zip ps qs \<rightarrow>\<^sub>h (p, q) |f ps p qs q. f ps \<rightarrow> p \<in> ta_rules A \<and> f qs \<rightarrow> q \<in> ta_rules B \<and> length ps = length qs} \<union>
    {[(p, q)] \<rightarrow>\<^sub>h (p', q) |p p' q. (p, p') \<in> ta_eps A} \<union>
    {[(p, q)] \<rightarrow>\<^sub>h (p, q') |p q q'. (q, q') \<in> ta_eps B}"

locale \<Delta>\<^sub>\<epsilon>'_horn =
  fixes A :: "('p, 'f) ta" and B :: "('q, 'f) ta"
begin

sublocale horn "\<Delta>\<^sub>\<epsilon>'_rules A B" .

lemma \<Delta>\<^sub>\<epsilon>'_sound:
  "\<Delta>\<^sub>\<epsilon>' A B = saturate"
proof (intro set_eqI iffI, goal_cases lr rl)
  case (lr x) obtain p q where x: "x = (p, q)" by (cases x)
  show ?case using lr unfolding x
  proof (induct)
    case (\<Delta>\<^sub>\<epsilon>'_cong f ps p qs q) show ?case
      apply (intro infer[of "zip ps qs" "(p, q)"])
      subgoal using \<Delta>\<^sub>\<epsilon>'_cong(1-3) by (auto simp: \<Delta>\<^sub>\<epsilon>'_rules_def)
      subgoal using \<Delta>\<^sub>\<epsilon>'_cong(3,5) by (auto simp: zip_nth_conv)
      done
  next
    case (\<Delta>\<^sub>\<epsilon>'_eps1 p p' q) then show ?case
      by (intro infer[of "[(p, q)]" "(p', q)"]) (auto simp: \<Delta>\<^sub>\<epsilon>'_rules_def)
  next
    case (\<Delta>\<^sub>\<epsilon>'_eps2 q q' p) then show ?case
      by (intro infer[of "[(p, q)]" "(p, q')"]) (auto simp: \<Delta>\<^sub>\<epsilon>'_rules_def)
  qed
next
  case (rl x) obtain p q where x: "x = (p, q)" by (cases x)
  show ?case using rl unfolding x
  proof (induct)
    case (infer as a) then show ?case
      using \<Delta>\<^sub>\<epsilon>'_cong[of _ "map fst as" "fst a" A "map snd as" "snd a" B]
        \<Delta>\<^sub>\<epsilon>'_eps1[of _ "fst a" A "snd a" B] \<Delta>\<^sub>\<epsilon>'_eps2[of _ "snd a" B "fst a" A]
      by (auto simp: \<Delta>\<^sub>\<epsilon>'_rules_def)
  qed
qed

end

definition \<Delta>\<^sub>\<epsilon>'_infer0 where
  "\<Delta>\<^sub>\<epsilon>'_infer0 A B = map (map_prod r_rhs r_rhs) (
     filter (\<lambda>(ra, rb). case (ra, rb) of (f ps \<rightarrow> p, g qs \<rightarrow> q) \<Rightarrow> f = g \<and> ps = [] \<and> qs = []) (List.product (tal_rules A) (tal_rules B)))"

definition \<Delta>\<^sub>\<epsilon>'_infer1 :: "('p, 'f) ta_list \<Rightarrow> ('q, 'f) ta_list \<Rightarrow> 'p \<times> 'q \<Rightarrow> ('p \<times> 'q) list \<Rightarrow> ('p \<times> 'q) list" where
  "\<Delta>\<^sub>\<epsilon>'_infer1 A B pq bs =
    map (map_prod r_rhs r_rhs) (filter (\<lambda>(ra, rb). case (ra, rb) of (f ps \<rightarrow> p, g qs \<rightarrow> q) \<Rightarrow> f = g \<and> length ps = length qs \<and> (fst pq, snd pq) \<in> set (zip ps qs) \<and> set (zip ps qs) \<subseteq> set ((fst pq, snd pq) # bs)) (List.product (tal_rules A) (tal_rules B))) @
    map (\<lambda>(p, p'). (p', snd pq)) (filter (\<lambda>(p, p') \<Rightarrow> p = fst pq) (tal_eps A)) @
    map (\<lambda>(q, q'). (fst pq, q')) (filter (\<lambda>(q, q') \<Rightarrow> q = snd pq) (tal_eps B))"

locale \<Delta>\<^sub>\<epsilon>'_list =
  fixes A :: "('p, 'f) ta_list" and B :: "('q, 'f) ta_list"
begin

sublocale \<Delta>\<^sub>\<epsilon>'_horn "ta_of A" "ta_of B" .

sublocale horn_list "\<Delta>\<^sub>\<epsilon>'_rules (ta_of A) (ta_of B)" "\<Delta>\<^sub>\<epsilon>'_infer0 A B" "\<Delta>\<^sub>\<epsilon>'_infer1 A B"
  (* TODO: clean up messy and brittle proof *)
  apply (unfold_locales)
  unfolding \<Delta>\<^sub>\<epsilon>'_rules_def horn_infer0_union horn_infer1_union
  unfolding horn.infer0_def horn.infer1_def \<Delta>\<^sub>\<epsilon>'_infer0_def \<Delta>\<^sub>\<epsilon>'_infer1_def set_append Un_assoc[symmetric]
  subgoal
    apply (subst HOL.trans[OF Un_empty_right Un_empty_right, symmetric, of "set _"])
    apply (intro arg_cong2[of _ _ _ _ "op \<union>"])
    subgoal
      apply (auto cong: conj_cong simp del: ex_simps(1) simp: ex_simps(1)[symmetric] simp: r_rhs_def split: ta_rule.splits)
      subgoal for f ps p qs q by (rule image_eqI[where x = "(f ps \<rightarrow> p, f qs \<rightarrow> q)"]) (cases ps; cases qs, auto)+
      subgoal for f p q by (auto intro!: exI[of _ f] exI[of _ "[]"])
      done
    subgoal by auto
    subgoal by auto
    done
  subgoal for x bs
    apply (cases x)
    apply (intro arg_cong2[of _ _ _ _ "op \<union>"])
    subgoal for p' q'
      apply (auto cong: conj_cong simp del: ex_simps(1) simp: ex_simps(1)[symmetric] simp: r_rhs_def[abs_def])
      subgoal for f ps p qs q by (rule image_eqI[where x = "(f ps \<rightarrow> p, f qs \<rightarrow> q)"]) auto
      subgoal by (auto split: ta_rule.splits)
      done
    subgoal by (auto cong: conj_cong simp del: ex_simps(1) simp: ex_simps(1)[symmetric])
    subgoal by (auto cong: conj_cong simp del: ex_simps(1) simp: ex_simps(1)[symmetric])
    done
  done

lemmas saturate_impl_sound = saturate_impl_sound
lemmas saturate_impl_complete = saturate_impl_complete

end

definition \<Delta>\<^sub>\<epsilon>'_impl :: "('p, 'f) ta_list \<Rightarrow> ('q, 'f) ta_list \<Rightarrow> ('p \<times> 'q) list option" where
  "\<Delta>\<^sub>\<epsilon>'_impl A B = horn_list_impl.saturate_impl (\<Delta>\<^sub>\<epsilon>'_infer0 A B) (\<Delta>\<^sub>\<epsilon>'_infer1 A B)"

lemma \<Delta>\<^sub>\<epsilon>'_impl_sound:
  "\<Delta>\<^sub>\<epsilon>'_impl A B = Some xs \<Longrightarrow> set xs = \<Delta>\<^sub>\<epsilon>' (ta_of A) (ta_of B)"
  using \<Delta>\<^sub>\<epsilon>'_list.saturate_impl_sound unfolding \<Delta>\<^sub>\<epsilon>'_impl_def \<Delta>\<^sub>\<epsilon>'_horn.\<Delta>\<^sub>\<epsilon>'_sound .

lemma \<Delta>\<^sub>\<epsilon>'_impl_complete:
  fixes A B :: "('q, 'f) ta_list"
  shows "\<Delta>\<^sub>\<epsilon>'_impl A B \<noteq> None"
proof -
  have *: "\<Delta>\<^sub>\<epsilon> (ta_of A) (ta_of B) \<subseteq> ta_states (ta_of A) \<times> ta_states (ta_of B)"
    by (auto simp: \<Delta>\<^sub>\<epsilon>_def ta_of_def subsetD[OF ta_res_states])
  have "finite (\<Delta>\<^sub>\<epsilon> (ta_of A) (ta_of B))"
    by (cases A; cases B; rule finite_subset[OF *])
      (auto simp: ta_of_def ta_states_def r_states_def intro!: finite_cartesian_product)
  then show ?thesis unfolding \<Delta>\<^sub>\<epsilon>'_impl_def
    by (intro \<Delta>\<^sub>\<epsilon>'_list.saturate_impl_complete) (auto simp: \<Delta>\<^sub>\<epsilon>'_horn.\<Delta>\<^sub>\<epsilon>'_sound \<Delta>\<^sub>\<epsilon>_eq_\<Delta>\<^sub>\<epsilon>')
qed

definition GTT_comp_impl :: "('p, 'f) gtt_list \<Rightarrow> ('p, 'f) gtt_list \<Rightarrow> (_, 'f) gtt_list option" where
  "GTT_comp_impl G1 G2 = (case (\<Delta>\<^sub>\<epsilon>'_impl (snd G1) (fst G2), \<Delta>\<^sub>\<epsilon>'_impl (fst G2) (snd G1)) of
     (Some xs, Some ys) \<Rightarrow> Some
     (ta_list [] (tal_rules (fst G1) @ tal_rules (fst G2)) (tal_eps (fst G1) @ tal_eps (fst G2) @ xs),
      ta_list [] (tal_rules (snd G1) @ tal_rules (snd G2)) (tal_eps (snd G1) @ tal_eps (snd G2) @ ys))
     | _ \<Rightarrow> None
     )"

lemma GTT_comp_impl_sound:
  "GTT_comp_impl G1 G2 = Some G \<Longrightarrow> gtt_of G = GTT_comp (gtt_of G1) (gtt_of G2)"
  by (auto simp: GTT_comp_def GTT_comp_impl_def \<Delta>\<^sub>\<epsilon>_eq_\<Delta>\<^sub>\<epsilon>' split: option.splits dest!: \<Delta>\<^sub>\<epsilon>'_impl_sound
   intro!: ta_of_eq_ta_makeI)

lemma GTT_comp_impl_complete:
  "GTT_comp_impl G1 G2 \<noteq> None"
  using \<Delta>\<^sub>\<epsilon>'_impl_complete[of "snd G1" "fst G2"] \<Delta>\<^sub>\<epsilon>'_impl_complete[of "fst G2" "snd G1"]
  by (auto simp: GTT_comp_impl_def)

(*
export_code GTT_comp_impl checking SML
export_code GTT_comp_impl in Haskell
*)

end