theory GTT_Transitive_Closure
  imports GTT_Compose Basic_Utils
begin

abbreviation "step_\<Delta>\<^sub>\<epsilon> \<A> \<B> \<equiv>
  (ta.make {} (ta_rules \<A>) (ta_eps \<A> \<union> \<Delta>\<^sub>\<epsilon> \<B> \<A>),
   ta.make {} (ta_rules \<B>) (ta_eps \<B> \<union> \<Delta>\<^sub>\<epsilon> \<A> \<B>))"

primrec make_\<Delta>\<^sub>\<epsilon> :: "('q, 'f) ta \<Rightarrow> ('q, 'f) ta \<Rightarrow> nat \<Rightarrow> ('q, 'f) ta \<times> ('q, 'f) ta" where
  "make_\<Delta>\<^sub>\<epsilon> \<A> \<B> 0 = (\<A>, \<B>)"
| "make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n) = (let (\<A>\<^sub>n, \<B>\<^sub>n) = make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n in step_\<Delta>\<^sub>\<epsilon> \<A>\<^sub>n \<B>\<^sub>n)"

abbreviation "states \<A> \<B> \<equiv> ta_states \<A> \<union> ta_states \<B>"
abbreviation "\<epsilon>_ext1 \<A> \<B> n \<equiv> ta_eps (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n))"
abbreviation "\<epsilon>_ext2 \<A> \<B> n \<equiv> ta_eps (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n))"

lemma \<Delta>\<^sub>\<epsilon>_subset1: "\<Delta>\<^sub>\<epsilon> \<A> \<B> \<subseteq> ta_states \<A> \<times> ta_states \<B>"
  unfolding \<Delta>\<^sub>\<epsilon>_def using ta_res_states by blast

lemma \<Delta>\<^sub>\<epsilon>_subset:
  assumes "ta_states \<A>' \<subseteq> states \<A> \<B>" "ta_states \<B>' \<subseteq> states \<A> \<B>"
  shows "\<Delta>\<^sub>\<epsilon> \<A>' \<B>' \<subseteq> states \<A> \<B> \<times> states \<A> \<B>"
  using assms \<Delta>\<^sub>\<epsilon>_subset1 by blast

lemma \<Delta>\<^sub>\<epsilon>_states:
  assumes "(q, q') \<in> \<Delta>\<^sub>\<epsilon> \<A> \<B>"
  shows "q \<in> states \<A> \<B>" "q' \<in> states \<A> \<B>"
  using assms \<Delta>\<^sub>\<epsilon>_subset1 by blast+

lemma make_\<Delta>\<^sub>\<epsilon>_ta_rules':
  "ta_rules (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n))) = ta_rules (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n))"
  "ta_rules (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n))) = ta_rules (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n))"
  unfolding ta.make_def make_\<Delta>\<^sub>\<epsilon>.simps
  by (metis (no_types, lifting) case_prod_unfold ext_inject fst_conv snd_conv surjective)+

lemma make_\<Delta>\<^sub>\<epsilon>_ta_rules:
  "ta_rules (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n)) = ta_rules (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> 0))"
  "ta_rules (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n)) = ta_rules (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> 0))"
  using make_\<Delta>\<^sub>\<epsilon>_ta_rules' by (induction n) fast+

lemma make_\<Delta>\<^sub>\<epsilon>_ta_eps':
  "\<epsilon>_ext1 \<A> \<B> n \<subseteq> \<epsilon>_ext1 \<A> \<B> (Suc n)" "\<epsilon>_ext2 \<A> \<B> n \<subseteq> \<epsilon>_ext2 \<A> \<B> (Suc n)"
  unfolding ta.make_def make_\<Delta>\<^sub>\<epsilon>.simps using fst_conv snd_conv
  by (metis (no_types, lifting) Un_upper1 case_prod_beta' select_convs(3))+

lemma make_\<Delta>\<^sub>\<epsilon>_ta_eps:
  assumes "m \<ge> n"
  shows "\<epsilon>_ext1 \<A> \<B> n \<subseteq> \<epsilon>_ext1 \<A> \<B> m" "\<epsilon>_ext2 \<A> \<B> n \<subseteq> \<epsilon>_ext2 \<A> \<B> m"
  using assms make_\<Delta>\<^sub>\<epsilon>_ta_eps' fst_conv snd_conv le_Suc_eq
  by (induction m) fastforce+

lemma subset_states_compose:
  assumes "\<Union> (r_states ` ta_rules \<A>) \<subseteq> Q" "ta_eps \<A> \<subseteq> Q \<times> Q" "ta_final \<A> \<subseteq> Q"
  shows "ta_states \<A> \<subseteq> Q"
  using assms unfolding ta_states_def by blast

lemma make_\<Delta>\<^sub>\<epsilon>_states_subset:
  "ta_states (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n)) \<subseteq> states \<A> \<B> \<and>
   ta_states (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n)) \<subseteq> states \<A> \<B>"
  unfolding make_\<Delta>\<^sub>\<epsilon>.simps ta.make_def
proof (induction n)
  case (Suc n)
  obtain \<A>\<^sub>n \<B>\<^sub>n where nth: "make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n = (\<A>\<^sub>n, \<B>\<^sub>n)" by fastforce
  hence "ta_eps \<A>\<^sub>n \<subseteq> states \<A> \<B> \<times> states \<A> \<B>" "ta_eps \<B>\<^sub>n \<subseteq> states \<A> \<B> \<times> states \<A> \<B>"
    using Suc by (auto simp: ta_states_def)
  moreover have "\<Delta>\<^sub>\<epsilon> \<B>\<^sub>n \<A>\<^sub>n \<subseteq> states \<A> \<B> \<times> states \<A> \<B>"
                "\<Delta>\<^sub>\<epsilon> \<A>\<^sub>n \<B>\<^sub>n \<subseteq> states \<A> \<B> \<times> states \<A> \<B>"
    using Suc by (simp_all add: \<Delta>\<^sub>\<epsilon>_subset nth)
  ultimately have eps: "ta_eps (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n))) \<subseteq> states \<A> \<B> \<times> states \<A> \<B>"
                       "ta_eps (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n))) \<subseteq> states \<A> \<B> \<times> states \<A> \<B>"
    unfolding make_\<Delta>\<^sub>\<epsilon>.simps ta.make_def nth by auto
  have final: "ta_final (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n))) \<subseteq> states \<A> \<B>"
               "ta_final (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n))) \<subseteq> states \<A> \<B>"
    by (auto simp: ta.make_def nth)
  have "\<Union> (r_states ` ta_rules \<A>\<^sub>n) \<subseteq> states \<A> \<B>"
       "\<Union> (r_states ` ta_rules \<B>\<^sub>n) \<subseteq> states \<A> \<B>"
    using make_\<Delta>\<^sub>\<epsilon>_ta_rules[of \<A> \<B> n] nth
    unfolding make_\<Delta>\<^sub>\<epsilon>.simps fst_conv snd_conv ta_states_def by auto
  hence rules: "\<Union> (r_states ` ta_rules (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n)))) \<subseteq> states \<A> \<B>"
               "\<Union> (r_states ` ta_rules (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n)))) \<subseteq> states \<A> \<B>"
    using nth unfolding make_\<Delta>\<^sub>\<epsilon>.simps ta.make_def by auto
  show ?case using subset_states_compose[OF rules(1) eps(1) final(1)]
    subset_states_compose[OF rules(2) eps(2) final(2)] by blast
qed simp

lemma \<epsilon>_ext1_subset: "\<epsilon>_ext1 \<A> \<B> n \<subseteq> states \<A> \<B> \<times> states \<A> \<B>"
  using make_\<Delta>\<^sub>\<epsilon>_states_subset[of \<A> \<B> n] unfolding ta_states_def by auto

lemma \<epsilon>_ext2_subset: "\<epsilon>_ext2 \<A> \<B> n \<subseteq> states \<A> \<B> \<times> states \<A> \<B>"
  using make_\<Delta>\<^sub>\<epsilon>_states_subset[of \<A> \<B> n] unfolding ta_states_def by auto

text \<open>The following lemma shows that there exists a
      saturated GTT obtained by iteratively adding \<epsilon>-transitions.\<close>
lemma trans_closure_fixpoint:
  assumes "finite (ta_states \<A>)" "finite (ta_states \<B>)"
  shows "\<exists>N. \<forall>n\<ge>N. make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n = make_\<Delta>\<^sub>\<epsilon> \<A> \<B> N"
proof -
  define S where "S = states \<A> \<B> \<times> states \<A> \<B>"
  have "finite S" using assms finite_subset S_def by blast
  have sub: "\<forall>n. \<epsilon>_ext1 \<A> \<B> n \<subseteq> S \<and> \<epsilon>_ext2 \<A> \<B> n \<subseteq> S"
    using \<epsilon>_ext1_subset \<epsilon>_ext2_subset S_def by blast
  have "\<exists>n. \<forall>m\<ge>n. (\<epsilon>_ext1 \<A> \<B> m) = (\<epsilon>_ext1 \<A> \<B> n) \<and>
                   (\<epsilon>_ext2 \<A> \<B> m) = (\<epsilon>_ext2 \<A> \<B> n)"
    using inc_set_fixpoint2[OF _ sub \<open>finite S\<close>] make_\<Delta>\<^sub>\<epsilon>_ta_eps by fast
  then obtain n where "\<forall>m\<ge>n. (\<epsilon>_ext1 \<A> \<B> m) = (\<epsilon>_ext1 \<A> \<B> n) \<and>
                              (\<epsilon>_ext2 \<A> \<B> m) = (\<epsilon>_ext2 \<A> \<B> n)" by blast
  from this this[THEN spec[of _ "Suc n"]] have
    "\<forall>m\<ge>Suc n. (\<epsilon>_ext1 \<A> \<B> m) = (\<epsilon>_ext1 \<A> \<B> (Suc n)) \<and> (\<epsilon>_ext2 \<A> \<B> m) = (\<epsilon>_ext2 \<A> \<B> (Suc n))"
    using Suc_leD  le_SucI[OF le_refl] by blast
  moreover have "\<forall>m\<ge>Suc n. ta_final (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> m)) = ta_final (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n))) \<and>
                           ta_final (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> m)) = ta_final (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n)))"
    by (auto split: prod.splits simp: ta.make_def dest: Suc_le_D)
  moreover have "\<forall>m\<ge>Suc n. ta_rules (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> m)) = ta_rules (fst (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n))) \<and>
                        ta_rules (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> m)) = ta_rules (snd (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n)))"
    using make_\<Delta>\<^sub>\<epsilon>_ta_rules[of \<A> \<B>]
    by (auto split: prod.splits simp del: make_\<Delta>\<^sub>\<epsilon>.simps) blast+
  ultimately have "\<forall>m\<ge>Suc n. make_\<Delta>\<^sub>\<epsilon> \<A> \<B> m = make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n)"
    using prod_eqI ta.equality unfolding make_\<Delta>\<^sub>\<epsilon>.simps ta.make_def
    by (metis (mono_tags, lifting) old.unit.exhaust)
  thus ?thesis by blast
qed

definition "least_fp_N \<A> \<B> \<equiv> LEAST N. \<forall>n\<ge>N. make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n = make_\<Delta>\<^sub>\<epsilon> \<A> \<B> N"

definition \<G>\<^sub>N :: "('q, 'f) ta \<Rightarrow> ('q, 'f) ta \<Rightarrow> ('q, 'f) ta \<times> ('q, 'f) ta" where
  "\<G>\<^sub>N \<A> \<B> = make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (least_fp_N \<A> \<B>)"

lemma \<G>\<^sub>N_is_fixpoint:
  assumes "finite (ta_states \<A>)" "finite (ta_states \<B>)"
  shows "step_\<Delta>\<^sub>\<epsilon> (fst (\<G>\<^sub>N \<A> \<B>)) (snd (\<G>\<^sub>N \<A> \<B>)) = \<G>\<^sub>N \<A> \<B>"
  using LeastI_ex[OF trans_closure_fixpoint[OF assms], THEN spec[of _ "Suc (least_fp_N \<A> \<B>)"]]
  by (cases "make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (least_fp_N \<A> \<B>)") (auto simp: \<G>\<^sub>N_def least_fp_N_def)

lemma \<G>\<^sub>N_GTT_comp:
  assumes "finite (ta_states \<A>)" "finite (ta_states \<B>)"
  shows "GTT_comp (\<G>\<^sub>N \<A> \<B>) (\<G>\<^sub>N \<A> \<B>) = \<G>\<^sub>N \<A> \<B>"
  using \<G>\<^sub>N_is_fixpoint[OF assms] by (auto simp: GTT_comp_def)

lemma \<G>\<^sub>N_GTT_trans:
  assumes "finite (ta_states \<A>)" "finite (ta_states \<B>)"
    "(s, u) \<in> gtt_lang (\<G>\<^sub>N \<A> \<B>)" "(u, t) \<in> gtt_lang (\<G>\<^sub>N \<A> \<B>)"
  shows "(s, t) \<in> gtt_lang (\<G>\<^sub>N \<A> \<B>)"
  using gtt_comp_lang_complete[of "\<G>\<^sub>N \<A> \<B>" "\<G>\<^sub>N \<A> \<B>"] assms(3,4)
  by (auto simp: \<G>\<^sub>N_GTT_comp[OF assms(1,2)])

lemma \<G>\<^sub>N_GTT_step:
  assumes "(s, t) \<in> gtt_lang (\<A>, \<B>)"
  shows "(s, t) \<in> gtt_lang (\<G>\<^sub>N \<A> \<B>)"
  using make_\<Delta>\<^sub>\<epsilon>_ta_eps[of 0 "least_fp_N \<A> \<B>", of \<A> \<B>, THEN subsetD] assms(1)
  by (auto intro!: gtt_accept_mono[of "(\<A>, \<B>)"] simp: \<G>\<^sub>N_def make_\<Delta>\<^sub>\<epsilon>_ta_rules)

lemma \<G>\<^sub>N_complete:
  assumes "finite (ta_states \<A>)" "finite (ta_states \<B>)"
  shows "Restr ((gtt_lang (\<A>, \<B>))\<^sup>*) {t. ground t} \<subseteq> gtt_lang (\<G>\<^sub>N \<A> \<B>)"
  unfolding rtrancl_trancl_reflcl Int_Un_distrib2
proof (intro Un_least, goal_cases)
  case 1
  have "gtt_lang (\<A>, \<B>) \<subseteq> gtt_lang (\<G>\<^sub>N \<A> \<B>)" using \<G>\<^sub>N_GTT_step[of _ _ \<A> \<B>] by auto
  note trancl_mono[OF _ this]
  moreover have "(gtt_lang (\<G>\<^sub>N \<A> \<B>))\<^sup>+ = gtt_lang (\<G>\<^sub>N \<A> \<B>)" using \<G>\<^sub>N_GTT_trans[OF assms]
    by (intro trancl_id transI)
  ultimately show ?case by blast
next
  case 2 then show ?case using gtt_accept.refl by blast
qed

lemma make_\<Delta>\<^sub>\<epsilon>_by_GTT_comp:
  "make_\<Delta>\<^sub>\<epsilon> \<A> \<B> (Suc n) = GTT_comp (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n) (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n)"
  by (cases "make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n") (auto simp: GTT_comp_def)

lemma gtt_lang_ground_ctxt_closed:
  assumes "ground_ctxt C" "(s, t) \<in> gtt_lang \<G>"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> gtt_lang \<G>"
proof -
  { fix i s and ss1 ss2 :: "('f, 'v) term list" assume "i < Suc (length ss1 + length ss2)"
    then have "fill_holes ((map mctxt_of_term ss1 @ MHole # map mctxt_of_term ss2) ! i)
     ((replicate (length ss1) [] @ [s] # replicate (length ss2) []) ! i) = (ss1 @ s # ss2) ! i"
      by (cases "i = length ss1") (auto simp del: upt_Suc simp: nth_append) (* slow *)
  } note [simp] = this
  have *: "fill_holes (MFun f (map mctxt_of_term ss1 @ MHole # map mctxt_of_term ss2)) [s] = Fun f (ss1 @ s # ss2)"
    for f ss1 s ss2 by (auto simp del: upt_Suc intro!: nth_equalityI)
  then have "ctxt.closed {(s, t). gtt_accept \<G> s t}"
    using accept'_closed_ctxt[of "[_]" "[_]" \<G> "MFun _ ((map mctxt_of_term _) @ MHole # (map mctxt_of_term _))"]
    unfolding * by (auto simp: gtt_accept_equiv intro!: one_imp_ctxt_closed)
  then show ?thesis using assms by auto
qed

lemma gtt_langs_ground_ctxt_closed:
  assumes "ground_ctxt C" "(s, t) \<in> (gtt_lang \<G>)\<^sup>*"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (gtt_lang \<G>)\<^sup>*"
  using assms(2) gtt_lang_ground_ctxt_closed[OF assms(1), of _ _ \<G>]
  by (induct rule: rtrancl_induct) (auto intro: rtrancl.intros(2))

(* move *)
lemma ground_ctxt_closed_imp_ground_mctxt_closed:
  assumes "\<And>C s t. ground_ctxt C \<Longrightarrow> (s, t) \<in> R\<^sup>* \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> R\<^sup>*"
    "length ss = num_holes C" "length ts = num_holes C" "ground_mctxt C"
    "\<And>i. i < num_holes C \<Longrightarrow> (ss ! i, ts ! i) \<in> Restr (R\<^sup>*) {t. ground t}"
  shows "(fill_holes C ss, fill_holes C ts) \<in> R\<^sup>*"
proof -
  have "i \<le> num_holes C \<Longrightarrow> (fill_holes C ss, fill_holes C (take i ts @ drop i ss)) \<in> R\<^sup>*" for i
  proof (induct i)
    case (Suc i)
    have "num_holes C = Suc (length (take i ts) + length (drop (Suc i) ss))"
      using Suc(2) assms(2,3) by auto
    from fill_holes_ctxt_main[OF this]
    obtain D where [simp]: "\<And>s. D\<langle>s\<rangle> = fill_holes C (take i ts @ s # drop (Suc i) ss)" by metis
    have [simp]: "take i ts @ ss ! i # drop (Suc i) ss = take i ts @ drop i ss"
      "take i ts @ ts ! i # drop (Suc i) ss = take (Suc i) ts @ drop (Suc i) ss"
      using Suc(2) assms(2,3)
      by (simp_all add: Cons_nth_drop_Suc Suc.prems Suc_le_lessD assms(3) take_Suc_conv_app_nth)
    have "ground D\<langle>ss ! i\<rangle>" using assms(2-5) by (auto simp: ground_fill_holes in_set_conv_nth)
    then have "ground_ctxt D" by (metis ground_ctxt_apply)
    then show ?case using Suc(1,2) assms(1)[of D "ss ! i" "ts ! i"] assms(5) by auto
  qed auto
  from this[of "num_holes C"] show ?thesis using assms(2,3) by auto
qed

(* move *)
lemma partition_by_singles:
  "length xs = n \<Longrightarrow> partition_by xs (replicate n (Suc 0)) = map (\<lambda>x. [x]) xs"
  by (induct xs arbitrary: n) auto

(* move *)
lemma map_nth':
  "length xs = n \<Longrightarrow> map (op ! xs) [0..<n] = xs"
  by (auto simp: map_nth)

(* move *)
lemma ground_ctxt_closed_imp_arg_closed:
  assumes "\<And>C s t. ground_ctxt C \<Longrightarrow> (s, t) \<in> R\<^sup>* \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> R\<^sup>*"
    "length ss = length ts"
    "\<And>i. i < length ts \<Longrightarrow> (ss ! i, ts ! i) \<in> Restr (R\<^sup>*) {t. ground t}"
  shows "(Fun f ss, Fun f ts) \<in> R\<^sup>*"
  using ground_ctxt_closed_imp_ground_mctxt_closed[of R ss "MFun f (replicate (length ts) MHole)" ts]
    assms(2,3) by (auto simp: assms(1) partition_by_singles map_nth' cong: list.map_cong_simp)

lemma rtrancl_gtt_lang_ground_mctxt_closed:
  assumes "length ss = num_holes C" "length ts = num_holes C" "ground_mctxt C"
    "\<forall>i < num_holes C. (ss ! i, ts ! i) \<in> Restr ((gtt_lang \<G>)\<^sup>*) {t. ground t}"
  shows "(fill_holes C ss, fill_holes C ts) \<in> (gtt_lang \<G>)\<^sup>*"
  using assms(4) ground_ctxt_closed_imp_ground_mctxt_closed[OF _ assms(1,2,3), of "gtt_lang \<G>"]
  by (auto simp: gtt_langs_ground_ctxt_closed)

lemma rtrancl_gtt_lang_arg_closed:
  assumes "length ss = length ts" "\<forall>i < length ts. (ss ! i, ts ! i) \<in> Restr ((gtt_lang \<G>)\<^sup>*) {t. ground t}"
  shows "(Fun f ss, Fun f ts) \<in> (gtt_lang \<G>)\<^sup>*"
  using assms(2) ground_ctxt_closed_imp_arg_closed[OF _ assms(1), of "gtt_lang \<G>"]
  by (auto simp: gtt_langs_ground_ctxt_closed)

lemma GTT_comp_\<G>_\<G>_in_rtrancl:
  "gtt_lang (GTT_comp \<G> \<G>) \<subseteq> (gtt_lang \<G>)\<^sup>*"
proof -
  { fix q t \<G> assume "q \<in> ta_res (fst (GTT_comp \<G> \<G>)) t" "ground t"
    then have "\<exists>s. ground s \<and> (t, s) \<in> (gtt_lang \<G>)\<^sup>* \<and> q \<in> ta_res (fst \<G>) s"
    proof (induct t arbitrary: q)
      case (Fun f ts)
      obtain qs q' where q':  "f qs \<rightarrow> q' \<in> ta_rules (fst (GTT_comp \<G> \<G>))"
        "(q', q) \<in> (ta_eps (fst (GTT_comp \<G> \<G>)))\<^sup>*"  "length qs = length ts"
        "\<And>i. i<length ts \<Longrightarrow> qs ! i \<in> ta_res (fst (GTT_comp \<G> \<G>)) (ts ! i)"
        using Fun(2) by auto
      { fix i assume i: "i < length ts"
        then have "\<exists>s. ground s \<and> (ts ! i, s) \<in> (gtt_lang \<G>)\<^sup>* \<and> qs ! i \<in> ta_res (fst \<G>) s"
          using Fun(1)[OF nth_mem q'(4), of i] Fun(3) by auto
      }
      then obtain ss where ss: "i < length ts \<Longrightarrow> ground (ss i) \<and> (ts ! i, ss i) \<in> (gtt_lang \<G>)\<^sup>* \<and>
        qs ! i \<in> ta_res (fst \<G>) (ss i)" for i by metis
      define s where "s = Fun f (map ss [0..<length ts])"
      have "ground s" using ss(1) by (auto simp: s_def)
      moreover have "(Fun f ts, s) \<in> (gtt_lang \<G>)\<^sup>*" using ss(1) Fun(3)
        by (auto intro!: rtrancl_gtt_lang_arg_closed simp: s_def)
      moreover have "q' \<in> ta_res (fst \<G>) s" using q'(1,3) ss by (auto simp: GTT_comp_def s_def)
      moreover show ?case using q'(2) calculation
      proof (induct arbitrary: s rule: converse_rtrancl_induct)
        case (step q' p)
        consider "(q', p) \<in> ta_eps (fst \<G>)" | "(q', p) \<in> \<Delta>\<^sub>\<epsilon> (snd \<G>) (fst \<G>)"
          using step(1) by (auto simp: GTT_comp_def)
        then show ?case
        proof (cases)
          case 1 then show ?thesis using step(3)[of s] step(4,5,6) by (auto intro: ta_res_eps)
        next
          case 2
          obtain s' where "(s, s') \<in> gtt_lang \<G>" "ground s'" "p \<in> ta_res (fst \<G>) s'"
            using 2 step(4,6) by (auto simp: \<Delta>\<^sub>\<epsilon>_def)
          then show ?thesis using step(5) rtrancl.rtrancl_into_rtrancl step.hyps(3) by force
        qed
      qed auto
    qed auto
  } note * = this
  have "gtt_accept' (GTT_comp \<G> \<G>) s t \<Longrightarrow> ground s \<Longrightarrow> ground t \<Longrightarrow> (s, t) \<in> (gtt_lang \<G>)\<^sup>*" for s t
  proof (induct rule: gtt_accept'.cases)
    case (mctxt ss ts C)
    { fix i assume i: "i < num_holes C"
      then obtain q where q: "q \<in> ta_res (fst (GTT_comp \<G> \<G>)) (ss ! i)" "q \<in> ta_res (snd (GTT_comp \<G> \<G>)) (ts ! i)"
        using mctxt(5) by (auto simp: mctxt(3,4))
      have [simp]: "ground (ss ! i)" "ground (ts ! i)" using i mctxt(3,4,6,7)
        by (auto simp add: ground_fill_holes mctxt(1,2))
      obtain s' where s': "ground s'" "q \<in> ta_res (fst \<G>) s'" "(ss ! i, s') \<in> (gtt_lang \<G>)\<^sup>*"
        using *[of q \<G> "ss ! i"] q(1) by auto
      obtain t' where t': "ground t'" "q \<in> ta_res (snd \<G>) t'" "(ts ! i, t') \<in> (gtt_lang (prod.swap \<G>))\<^sup>*"
        using *[of q "prod.swap \<G>" "ts ! i"] q(2) by auto
      have "(gtt_lang (prod.swap \<G>))\<inverse> = gtt_lang \<G>" by auto
      then have "(t', ts ! i) \<in> (gtt_lang \<G>)\<^sup>*" using converseI[OF t'(3)] unfolding converse_inward(1) by metis
      moreover have "(s', t') \<in> gtt_lang \<G>" using s'(1,2) t'(1,2) by auto
      ultimately have "(ss ! i, ts ! i) \<in> (gtt_lang \<G>)\<^sup>*" using s'(3)
        by (meson rtrancl.rtrancl_into_rtrancl rtrancl_trans)
    }
    then show ?case using mctxt(6,7)
      by (auto simp: ground_fill_holes mctxt(1-4) intro!: rtrancl_gtt_lang_ground_mctxt_closed)
  qed
  then show ?thesis by (auto simp: gtt_accept_equiv)
qed

lemma make_\<Delta>\<^sub>\<epsilon>_sound:
  "gtt_lang (make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n) \<subseteq> (gtt_lang (\<A>, \<B>))\<^sup>*"
proof (induct n)
  case (Suc n) show ?case using rtrancl_mono[OF Suc] GTT_comp_\<G>_\<G>_in_rtrancl[of "make_\<Delta>\<^sub>\<epsilon> \<A> \<B> n"]
    by (simp del: make_\<Delta>\<^sub>\<epsilon>.simps add: make_\<Delta>\<^sub>\<epsilon>_by_GTT_comp)
qed auto

lemma \<G>\<^sub>N_sound:
  "gtt_lang (\<G>\<^sub>N \<A> \<B>) \<subseteq> Restr ((gtt_lang (\<A>, \<B>))\<^sup>*) {t. ground t}"
  by (auto simp: \<G>\<^sub>N_def make_\<Delta>\<^sub>\<epsilon>_sound)

lemma \<G>\<^sub>N_lang:
  assumes "finite (ta_states \<A>)" "finite (ta_states \<B>)"
  shows "gtt_lang (\<G>\<^sub>N \<A> \<B>) = Restr ((gtt_lang (\<A>, \<B>))\<^sup>*) {t. ground t}"
  using \<G>\<^sub>N_sound \<G>\<^sub>N_complete[OF assms] by auto

lemma Restr_rtrancl_gtt_lang_eq_trancl_gtt_lang:
  "Restr ((gtt_lang (\<A>, \<B>))\<^sup>*) {t. ground t} = (gtt_lang (\<A>, \<B>))\<^sup>+"
  by (auto simp: rtrancl_trancl_reflcl simp del: reflcl_trancl dest: tranclD tranclD2)

abbreviation "GTT_rtrancl G \<equiv> \<G>\<^sub>N (fst G) (snd G)"

end