theory GTT_Transitive_Closure_Impl
  imports GTT_Compose_Impl GTT_Transitive_Closure
begin

(* Future plan: compute the new epsilon transitions in a single inference pass *)

inductive_set \<Delta>_trancl' :: "('q, 'f) ta \<Rightarrow> ('q, 'f) ta \<Rightarrow> ('q \<times> 'q) set" for A B where
  \<Delta>_cong: "f ps \<rightarrow> p \<in> ta_rules A \<Longrightarrow> f qs \<rightarrow> q \<in> ta_rules B \<Longrightarrow> length ps = length qs \<Longrightarrow>
   (\<And>i. i < length qs \<Longrightarrow> (ps ! i, qs ! i) \<in> \<Delta>_trancl' A B) \<Longrightarrow> (p, q) \<in> \<Delta>_trancl' A B"
| \<Delta>_eps1: "(p, p') \<in> ta_eps A \<Longrightarrow> (p, q) \<in> \<Delta>_trancl' A B \<Longrightarrow> (p', q) \<in> \<Delta>_trancl' A B"
| \<Delta>_eps2: "(q, q') \<in> ta_eps B \<Longrightarrow> (p, q) \<in> \<Delta>_trancl' A B \<Longrightarrow> (p, q') \<in> \<Delta>_trancl' A B"
| \<Delta>_trans: "(p, q) \<in> \<Delta>_trancl A B \<Longrightarrow> (q, r) \<in> \<Delta>_trancl' A B \<Longrightarrow> (p, r) \<in> \<Delta>_trancl' A B"

definition "GTT_rtrancl' G \<equiv>
  (ta.make {} (ta_rules (fst G)) (ta_eps (fst G) \<union> \<Delta>_trancl' (snd G) (fst G)),
   ta.make {} (ta_rules (snd G)) (ta_eps (snd G) \<union> \<Delta>_trancl' (fst G) (snd G)))"

lemma [code_unfold]:
  "GTT_rtrancl G = GTT_rtrancl' G"
  oops

(* But for now let's implement \<G>\<^sub>N as simply as possible *)

partial_function (tailrec) GTT_rtrancl_impl :: "('q, 'f) gtt_list \<Rightarrow> ('q, 'f) gtt_list option" where
  [code]: "GTT_rtrancl_impl G = (case GTT_comp_impl G G of
       Some G' \<Rightarrow> (let G'' = map_prod uniq_ta uniq_ta G' in
       (if gtt_of G = gtt_of G'' then Some G else GTT_rtrancl_impl G''))
     | None \<Rightarrow> None)"

lemma step_\<Delta>\<^sub>\<epsilon>_to_GTT_comp:
  "step_\<Delta>\<^sub>\<epsilon> A B = GTT_comp (A, B) (A, B)"
  by (auto simp: GTT_comp_def)

lemma make_\<Delta>\<^sub>\<epsilon>_GTT_comp_commute:
  "GTT_comp (make_\<Delta>\<^sub>\<epsilon> A B n) (make_\<Delta>\<^sub>\<epsilon> A B n) =
   make_\<Delta>\<^sub>\<epsilon> (fst (GTT_comp (A, B) (A, B))) (snd (GTT_comp (A, B) (A, B))) n"
  by (induct n) (simp_all add: step_\<Delta>\<^sub>\<epsilon>_to_GTT_comp)

lemma step_\<Delta>\<^sub>\<epsilon>_decrease:
  assumes "finite (ta_states A)" "finite (ta_states B)"
  shows "least_fp_N A B = 0 \<and> step_\<Delta>\<^sub>\<epsilon> A B = (A, B) \<or> least_fp_N A B = Suc (case_prod least_fp_N (step_\<Delta>\<^sub>\<epsilon> A B))"
proof -
  have *: "(\<forall>x \<ge> Suc m. P x) \<longleftrightarrow> (\<forall>x \<ge> m. P (Suc x))" for m P
    using Suc_le_D by blast
  show ?thesis
  apply (cases "least_fp_N A B")
  subgoal using spec[OF LeastI_ex[OF trans_closure_fixpoint[OF assms], folded least_fp_N_def], of 1]
    by simp
  subgoal for n
    using Least_Suc[of "\<lambda>n. \<forall>m \<ge> n. make_\<Delta>\<^sub>\<epsilon> A B m = make_\<Delta>\<^sub>\<epsilon> A B n", OF LeastI_ex[OF trans_closure_fixpoint[OF assms]]]
      not_less_Least[of 0 "\<lambda>n. \<forall>m \<ge> n. make_\<Delta>\<^sub>\<epsilon> A B m = make_\<Delta>\<^sub>\<epsilon> A B n"]
    unfolding least_fp_N_def * make_\<Delta>\<^sub>\<epsilon>.simps Let_def step_\<Delta>\<^sub>\<epsilon>_to_GTT_comp
    by (auto simp: case_prod_beta make_\<Delta>\<^sub>\<epsilon>_GTT_comp_commute least_fp_N_def[symmetric])
  done
qed

lemma finite_ta_states_ta_of:
  "finite (ta_states (ta_of A))"
  by (cases A) (auto simp: ta_of_def ta_states_def r_states_def)

lemma GTT_rtrancl_impl_sound:
  "GTT_rtrancl_impl G = Some G' \<Longrightarrow> gtt_of G' = GTT_rtrancl (gtt_of G)"
  unfolding \<G>\<^sub>N_def fst_map_prod snd_map_prod
proof (induct "least_fp_N (ta_of (fst G)) (ta_of (snd G))" arbitrary: G rule: less_induct)
  case less show ?case
  proof (cases "GTT_comp_impl G G")
    case None then show ?thesis using less(2) by (subst (asm) GTT_rtrancl_impl.simps) auto
  next
    case (Some G'')
    show ?thesis
      using less(1)[unfolded map_prod_def[abs_def] case_prod_beta, of "map_prod uniq_ta uniq_ta G''"] less(2)
        arg_cong[OF GTT_comp_impl_sound[OF Some], of fst, symmetric]
        arg_cong[OF GTT_comp_impl_sound[OF Some], of snd, symmetric]
      apply (subst (asm) GTT_rtrancl_impl.simps)
      apply (cases rule: disjE[OF
       step_\<Delta>\<^sub>\<epsilon>_decrease[OF finite_ta_states_ta_of finite_ta_states_ta_of, unfolded step_\<Delta>\<^sub>\<epsilon>_to_GTT_comp,
       of "fst G" "snd G", unfolded case_prod_beta]])
       apply (auto simp: Some map_prod_def case_prod_beta)[1]
      apply (simp only: make_\<Delta>\<^sub>\<epsilon>.simps step_\<Delta>\<^sub>\<epsilon>_to_GTT_comp)
      apply (auto simp: make_\<Delta>\<^sub>\<epsilon>_GTT_comp_commute Some Let_def map_prod_def case_prod_beta
        dest: GTT_comp_impl_sound split: if_splits)
      done
  qed
qed

(* export_code GTT_rtrancl_impl in Haskell *)

end